﻿Imports System.Linq
Imports System.Text
Imports System.Threading.Tasks
Imports System.Windows.Input
Imports Windows.System
Imports Windows.UI.Core
Imports Windows.UI.Xaml
Imports Windows.UI.Xaml.Controls
Imports Windows.UI.Xaml.Navigation
Imports Windows.UI.Input
Imports System
Imports System.Collections.Generic

Namespace WsaXamlCs.Common
    ''' <summary>
    ''' NavigationManager aids in navigation between pages.  It provides commands used to 
    ''' navigate back and forward as well as registers for standard mouse and keyboard 
    ''' shortcuts used to go back and forward.  In addition it integrates SuspensionManger
    ''' to handle process lifetime management and state management when navigating between
    ''' pages.
    ''' </summary>
    ''' <example>
    ''' To make use of NavigationManager, follow these two steps or
    ''' start with a BasicPage or any other Page item template other than BlankPage.
    ''' 
    ''' 1) Create an instance of the NaivgationHelper somewhere such as in the 
    '''     constructor for the page and register a callback for the LoadState and 
    '''     SaveState events.
    ''' <code>
    '''     public MyPage()
    '''     {
    '''         this.InitializeComponent();
    '''         var navigationHelper = new NavigationHelper(this);
    '''         this.navigationHelper.LoadState += navigationHelper_LoadState;
    '''         this.navigationHelper.SaveState += navigationHelper_SaveState;
    '''     }
    '''     
    '''     private async void navigationHelper_LoadState(object sender, LoadStateEventArgs e)
    '''     { }
    '''     private async void navigationHelper_SaveState(object sender, LoadStateEventArgs e)
    '''     { }
    ''' </code>
    ''' 
    ''' 2) Register the page to call into the NavigationManager whenever the page participates 
    '''     in navigation by overriding the <see cref="Windows.UI.Xaml.Controls.Page.OnNavigatedTo"/> 
    '''     and <see cref="Windows.UI.Xaml.Controls.Page.OnNavigatedFrom"/> events.
    ''' <code>
    '''     protected override void OnNavigatedTo(NavigationEventArgs e)
    '''     {
    '''         navigationHelper.OnNavigatedTo(e);
    '''     }
    '''     
    '''     protected override void OnNavigatedFrom(NavigationEventArgs e)
    '''     {
    '''         navigationHelper.OnNavigatedFrom(e);
    '''     }
    ''' </code>
    ''' </example>
    <Windows.Foundation.Metadata.WebHostHidden>
    Public Class NavigationHelper
        Inherits DependencyObject
        Private Property Page() As Page
        Private ReadOnly Property Frame() As Frame
            Get
                Return Me.Page.Frame
            End Get
        End Property

        ''' <summary>
        ''' Initializes a new instance of the <see cref="NavigationHelper"/> class.
        ''' </summary>
        ''' <param name="page">A reference to the current page used for navigation.  
        ''' This reference allows for frame manipulation and to ensure that keyboard 
        ''' navigation requests only occur when the page is occupying the entire window.</param>
        Public Sub New(ByVal page As Page)
            Me.Page = page

            ' When this page is part of the visual tree make two changes:
            ' 1) Map application view state to visual state for the page
            ' 2) Handle keyboard and mouse navigation requests
            AddHandler Me.Page.Loaded, Sub(sender, e)
                                           ' Keyboard and mouse navigation only apply when occupying the entire window
                                           If Me.Page.ActualHeight = Window.Current.Bounds.Height AndAlso Me.Page.ActualWidth = Window.Current.Bounds.Width Then
                                               ' Listen to the window directly so focus isn't required
                                               AddHandler Window.Current.CoreWindow.Dispatcher.AcceleratorKeyActivated, AddressOf CoreDispatcher_AcceleratorKeyActivated
                                               AddHandler Window.Current.CoreWindow.PointerPressed, AddressOf CoreWindow_PointerPressed
                                           End If
                                       End Sub

            ' Undo the same changes when the page is no longer visible
            AddHandler Me.Page.Unloaded, Sub(sender, e)
                                             RemoveHandler Window.Current.CoreWindow.Dispatcher.AcceleratorKeyActivated, AddressOf CoreDispatcher_AcceleratorKeyActivated
                                             RemoveHandler Window.Current.CoreWindow.PointerPressed, AddressOf CoreWindow_PointerPressed
                                         End Sub
        End Sub

#Region "Navigation support"

        Private _goBackCommand As RelayCommand
        Private _goForwardCommand As RelayCommand

        ''' <summary>
        ''' <see cref="RelayCommand"/> used to bind to the back Button's Command property
        ''' for navigating to the most recent item in back navigation history, if a Frame
        ''' manages its own navigation history.
        ''' 
        ''' The <see cref="RelayCommand"/> is set up to use the virtual method <see cref="GoBack"/>
        ''' as the Execute Action and <see cref="CanGoBack"/> for CanExecute.
        ''' </summary>
        Public Property GoBackCommand() As RelayCommand
            Get
                If _goBackCommand Is Nothing Then
                    _goBackCommand = New RelayCommand(Sub() Me.GoBack(), Function() Me.CanGoBack())
                End If
                Return _goBackCommand
            End Get
            Set(ByVal value As RelayCommand)
                _goBackCommand = value
            End Set
        End Property
        ''' <summary>
        ''' <see cref="RelayCommand"/> used for navigating to the most recent item in 
        ''' the forward navigation history, if a Frame manages its own navigation history.
        ''' 
        ''' The <see cref="RelayCommand"/> is set up to use the virtual method <see cref="GoForward"/>
        ''' as the Execute Action and <see cref="CanGoForward"/> for CanExecute.
        ''' </summary>
        Public ReadOnly Property GoForwardCommand() As RelayCommand
            Get
                If _goForwardCommand Is Nothing Then
                    _goForwardCommand = New RelayCommand(Sub() Me.GoForward(), Function() Me.CanGoForward())
                End If
                Return _goForwardCommand
            End Get
        End Property

        ''' <summary>
        ''' Virtual method used by the <see cref="GoBackCommand"/> property
        ''' to determine if the <see cref="Frame"/> can go back.
        ''' </summary>
        ''' <returns>
        ''' true if the <see cref="Frame"/> has at least one entry 
        ''' in the back navigation history.
        ''' </returns>
        Public Overridable Function CanGoBack() As Boolean
            Return Me.Frame IsNot Nothing AndAlso Me.Frame.CanGoBack
        End Function
        ''' <summary>
        ''' Virtual method used by the <see cref="GoForwardCommand"/> property
        ''' to determine if the <see cref="Frame"/> can go forward.
        ''' </summary>
        ''' <returns>
        ''' true if the <see cref="Frame"/> has at least one entry 
        ''' in the forward navigation history.
        ''' </returns>
        Public Overridable Function CanGoForward() As Boolean
            Return Me.Frame IsNot Nothing AndAlso Me.Frame.CanGoForward
        End Function

        ''' <summary>
        ''' Virtual method used by the <see cref="GoBackCommand"/> property
        ''' to invoke the <see cref="Windows.UI.Xaml.Controls.Frame.GoBack"/> method.
        ''' </summary>
        Public Overridable Sub GoBack()
            If Me.Frame IsNot Nothing AndAlso Me.Frame.CanGoBack Then
                Me.Frame.GoBack()
            End If
        End Sub
        ''' <summary>
        ''' Virtual method used by the <see cref="GoForwardCommand"/> property
        ''' to invoke the <see cref="Windows.UI.Xaml.Controls.Frame.GoForward"/> method.
        ''' </summary>
        Public Overridable Sub GoForward()
            If Me.Frame IsNot Nothing AndAlso Me.Frame.CanGoForward Then
                Me.Frame.GoForward()
            End If
        End Sub

        ''' <summary>
        ''' Invoked on every keystroke, including system keys such as Alt key combinations, when
        ''' this page is active and occupies the entire window.  Used to detect keyboard navigation
        ''' between pages even when the page itself doesn't have focus.
        ''' </summary>
        ''' <param name="sender">Instance that triggered the event.</param>
        ''' <param name="e">Event data describing the conditions that led to the event.</param>
        Private Sub CoreDispatcher_AcceleratorKeyActivated(ByVal sender As CoreDispatcher, ByVal e As AcceleratorKeyEventArgs)
            Dim virtualKey As VirtualKey = e.VirtualKey

            ' Only investigate further when Left, Right, or the dedicated Previous or Next keys
            ' are pressed
            If (e.EventType = CoreAcceleratorKeyEventType.SystemKeyDown OrElse e.EventType = CoreAcceleratorKeyEventType.KeyDown) AndAlso (virtualKey = virtualKey.Left OrElse virtualKey = virtualKey.Right OrElse CInt(virtualKey) = 166 OrElse CInt(virtualKey) = 167) Then
                Dim coreWindow As CoreWindow = Window.Current.CoreWindow
                Dim downState As CoreVirtualKeyStates = CoreVirtualKeyStates.Down
                Dim menuKey As Boolean = (coreWindow.GetKeyState(virtualKey.Menu) And downState) = downState
                Dim controlKey As Boolean = (coreWindow.GetKeyState(virtualKey.Control) And downState) = downState
                Dim shiftKey As Boolean = (coreWindow.GetKeyState(virtualKey.Shift) And downState) = downState
                Dim noModifiers As Boolean = (Not menuKey) AndAlso (Not controlKey) AndAlso Not shiftKey
                Dim onlyAlt As Boolean = menuKey AndAlso (Not controlKey) AndAlso Not shiftKey

                If (CInt(virtualKey) = 166 AndAlso noModifiers) OrElse (virtualKey = virtualKey.Left AndAlso onlyAlt) Then
                    ' When the previous key or Alt+Left are pressed navigate back
                    e.Handled = True
                    Me.GoBackCommand.Execute(Nothing)
                ElseIf (CInt(virtualKey) = 167 AndAlso noModifiers) OrElse (virtualKey = virtualKey.Right AndAlso onlyAlt) Then
                    ' When the next key or Alt+Right are pressed navigate forward
                    e.Handled = True
                    Me.GoForwardCommand.Execute(Nothing)
                End If
            End If
        End Sub

        ''' <summary>
        ''' Invoked on every mouse click, touch screen tap, or equivalent interaction when this
        ''' page is active and occupies the entire window.  Used to detect browser-style next and
        ''' previous mouse button clicks to navigate between pages.
        ''' </summary>
        ''' <param name="sender">Instance that triggered the event.</param>
        ''' <param name="e">Event data describing the conditions that led to the event.</param>
        Private Sub CoreWindow_PointerPressed(ByVal sender As CoreWindow, ByVal e As PointerEventArgs)
            Dim properties As PointerPointProperties = e.CurrentPoint.Properties

            ' Ignore button chords with the left, right, and middle buttons
            If properties.IsLeftButtonPressed OrElse properties.IsRightButtonPressed OrElse properties.IsMiddleButtonPressed Then
                Return
            End If

            ' If back or foward are pressed (but not both) navigate appropriately
            Dim backPressed As Boolean = properties.IsXButton1Pressed
            Dim forwardPressed As Boolean = properties.IsXButton2Pressed
            If backPressed Xor forwardPressed Then
                e.Handled = True
                If backPressed Then
                    Me.GoBackCommand.Execute(Nothing)
                End If
                If forwardPressed Then
                    Me.GoForwardCommand.Execute(Nothing)
                End If
            End If
        End Sub

#End Region

#Region "Process lifetime management"

        Private _pageKey As String

        ''' <summary>
        ''' Register this event on the current page to populate the page
        ''' with content passed during navigation as well as any saved
        ''' state provided when recreating a page from a prior session.
        ''' </summary>
        Public Event LoadState As LoadStateEventHandler
        ''' <summary>
        ''' Register this event on the current page to preserve
        ''' state associated with the current page in case the
        ''' application is suspended or the page is discarded from
        ''' the navigaqtion cache.
        ''' </summary>
        Public Event SaveState As SaveStateEventHandler

        ''' <summary>
        ''' Invoked when this page is about to be displayed in a Frame.  
        ''' This method calls <see cref="LoadState"/>, where all page specific
        ''' navigation and process lifetime management logic should be placed.
        ''' </summary>
        ''' <param name="e">Event data that describes how this page was reached.  The Parameter
        ''' property provides the group to be displayed.</param>
        Public Sub OnNavigatedTo(ByVal e As NavigationEventArgs)
            Dim frameState As Dictionary(Of String, Object) = SuspensionManager.SessionStateForFrame(Me.Frame)
            Me._pageKey = "Page-" & Me.Frame.BackStackDepth

            If e.NavigationMode = NavigationMode.New Then
                ' Clear existing state for forward navigation when adding a new page to the
                ' navigation stack
                Dim nextPageKey As String = Me._pageKey
                Dim nextPageIndex As Integer = Me.Frame.BackStackDepth
                Do While frameState.Remove(nextPageKey)
                    nextPageIndex += 1
                    nextPageKey = "Page-" & nextPageIndex
                Loop

                ' Pass the navigation parameter to the new page
                RaiseEvent LoadState(Me, New LoadStateEventArgs(e.Parameter, Nothing))
            Else
                ' Pass the navigation parameter and preserved page state to the page, using
                ' the same strategy for loading suspended state and recreating pages discarded
                ' from cache
                RaiseEvent LoadState(Me, New LoadStateEventArgs(e.Parameter, CType(frameState(Me._pageKey), Dictionary(Of String, Object))))
            End If
        End Sub

        ''' <summary>
        ''' Invoked when this page will no longer be displayed in a Frame.
        ''' This method calls <see cref="SaveState"/>, where all page specific
        ''' navigation and process lifetime management logic should be placed.
        ''' </summary>
        ''' <param name="e">Event data that describes how this page was reached.  The Parameter
        ''' property provides the group to be displayed.</param>
        Public Sub OnNavigatedFrom(ByVal e As NavigationEventArgs)
            Dim frameState As Dictionary(Of String, Object) = SuspensionManager.SessionStateForFrame(Me.Frame)
            Dim pageState As Dictionary(Of String, Object) = New Dictionary(Of String, Object)()
            RaiseEvent SaveState(Me, New SaveStateEventArgs(pageState))
            frameState(_pageKey) = pageState
        End Sub

#End Region
    End Class

    ''' <summary>
    ''' Represents the method that will handle the <see cref="NavigationHelper.LoadState"/>event
    ''' </summary>
    Public Delegate Sub LoadStateEventHandler(ByVal sender As Object, ByVal e As LoadStateEventArgs)
    ''' <summary>
    ''' Represents the method that will handle the <see cref="NavigationHelper.SaveState"/>event
    ''' </summary>
    Public Delegate Sub SaveStateEventHandler(ByVal sender As Object, ByVal e As SaveStateEventArgs)

    ''' <summary>
    ''' Class used to hold the event data required when a page attempts to load state.
    ''' </summary>
    Public Class LoadStateEventArgs
        Inherits EventArgs
        ''' <summary>
        ''' The parameter value passed to <see cref="Frame.Navigate(Type, Object)"/> 
        ''' when this page was initially requested.
        ''' </summary>
        Private privateNavigationParameter As Object
        Public Property NavigationParameter() As Object
            Get
                Return privateNavigationParameter
            End Get
            Private Set(ByVal value As Object)
                privateNavigationParameter = value
            End Set
        End Property
        ''' <summary>
        ''' A dictionary of state preserved by this page during an earlier
        ''' session.  This will be null the first time a page is visited.
        ''' </summary>
        Private privatePageState As Dictionary(Of String, Object)
        Public Property PageState() As Dictionary(Of String, Object)
            Get
                Return privatePageState
            End Get
            Private Set(ByVal value As Dictionary(Of String, Object))
                privatePageState = value
            End Set
        End Property

        ''' <summary>
        ''' Initializes a new instance of the <see cref="LoadStateEventArgs"/> class.
        ''' </summary>
        ''' <param name="navigationParameter">
        ''' The parameter value passed to <see cref="Frame.Navigate(Type, Object)"/> 
        ''' when this page was initially requested.
        ''' </param>
        ''' <param name="pageState">
        ''' A dictionary of state preserved by this page during an earlier
        ''' session.  This will be null the first time a page is visited.
        ''' </param>
        Public Sub New(ByVal navigationParameter As Object, ByVal pageState As Dictionary(Of String, Object))
            MyBase.New()
            Me.NavigationParameter = navigationParameter
            Me.PageState = pageState
        End Sub
    End Class
    ''' <summary>
    ''' Class used to hold the event data required when a page attempts to save state.
    ''' </summary>
    Public Class SaveStateEventArgs
        Inherits EventArgs
        ''' <summary>
        ''' An empty dictionary to be populated with serializable state.
        ''' </summary>
        Private privatePageState As Dictionary(Of String, Object)
        Public Property PageState() As Dictionary(Of String, Object)
            Get
                Return privatePageState
            End Get
            Private Set(ByVal value As Dictionary(Of String, Object))
                privatePageState = value
            End Set
        End Property

        ''' <summary>
        ''' Initializes a new instance of the <see cref="SaveStateEventArgs"/> class.
        ''' </summary>
        ''' <param name="pageState">An empty dictionary to be populated with serializable state.</param>
        Public Sub New(ByVal pageState As Dictionary(Of String, Object))
            MyBase.New()
            Me.PageState = pageState
        End Sub
    End Class
End Namespace
